using System;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace YahooMessenger
{
	public class YahooData
	{
		public bool Login;
		public string name;
		public string password;
		public int state;
		public string state_msg;
		public bool state_away;
		public int id;
		public bool AutoLogin;

		public YahooData() { Login = false; AutoLogin = false; }
		public YahooData(string _name, string _password)
		{
			name = _name;
			password = _password;
			Login = false;
			AutoLogin = false; 
		}
	}

	public class YahooParameter
	{
		public int type;
		public byte[] data;
		
		public YahooParameter() {}
		public YahooParameter(int _type, byte[] _data)
		{
			type = _type;
			data = _data;
		}
	}

	public class YahooPacket
	{
		public short service;
		public int status;
		public int id;
		public YahooParameter[] data;
		public YahooPacket() { data = new YahooParameter[0]; }
		public YahooPacket(short _service, int _status, int _id)
		{
			service = _service;
			status  = _status;
			id      = _id;
			data = new YahooParameter[0];
		}

		public int GetBodyLength()
		{
			int len = 0;
			string s;
			for(int i = 0; i < data.Length; i++)
			{
				s = data[i].type.ToString();
				len += s.Length;
				len += 2; // "C0 80"
				len += data[i].data.Length;
				len += 2; // "C0 80"
			}
			return len;
		}

		public byte[] Dump()
		{
			byte[] b = new byte[20 + GetBodyLength()];
			byte[] pad = new byte[2] {0xC0, 0x80};
			int pos = 0;
			pos = YahooUtility.PutDataToByteArray(b, pos, "YMSG");
			pos = YahooUtility.PutDataToByteArray(b, pos, IPAddress.HostToNetworkOrder((short)0x0009));
			pos = YahooUtility.PutDataToByteArray(b, pos, IPAddress.HostToNetworkOrder((short)0x0000));
			pos = YahooUtility.PutDataToByteArray(b, pos, IPAddress.HostToNetworkOrder((short)GetBodyLength()));
			pos = YahooUtility.PutDataToByteArray(b, pos, IPAddress.HostToNetworkOrder(service));
			pos = YahooUtility.PutDataToByteArray(b, pos, IPAddress.HostToNetworkOrder(status));
			pos = YahooUtility.PutDataToByteArray(b, pos, IPAddress.HostToNetworkOrder(id));
			
			for(int i = 0; i < data.Length; i++)
			{
				pos = YahooUtility.PutDataToByteArray(b, pos, data[i].type.ToString());
				pos = YahooUtility.PutDataToByteArray(b, pos, pad);
				pos = YahooUtility.PutDataToByteArray(b, pos, data[i].data);
				pos = YahooUtility.PutDataToByteArray(b, pos, pad);
			}
			return b;
		}

		public void Add(YahooParameter param)
		{
			int l = data.Length;
			YahooParameter[] newdata = new YahooParameter[l + 1];
			data.CopyTo(newdata, 0);
			newdata[l] = param;
			data = newdata;
		}

		public void Add(int _type, byte[] _data)
		{
			YahooParameter adddata = new YahooParameter(_type, _data);
			Add(adddata);
		}
	}

	#region e萔
	public enum YahooService 
	{
		LOGON = 1,
		LOGOFF,
		ISAWAY,
		ISBACK,
		IDLE, /* 5 (placemarker) */
		MESSAGE,
		IDACT,
		IDDEACT,
		MAILSTAT,
		USERSTAT, /* 0xa */
		NEWMAIL,
		CHATINVITE,
		CALENDAR,
		NEWPERSONALMAIL,
		NEWCONTACT,
		ADDIDENT, /* 0x10 */
		ADDIGNORE,
		PING,
		GOTGROUPRENAME, /* < 1, 36(old), 37(new) */
		SYSMESSAGE = 0x14,
		PASSTHROUGH2 = 0x16,
		CONFINVITE = 0x18,
		CONFLOGON,
		CONFDECLINE,
		CONFLOGOFF,
		CONFADDINVITE,
		CONFMSG,
		CHATLOGON,
		CHATLOGOFF,
		CHATMSG = 0x20,
		GAMELOGON = 0x28,
		GAMELOGOFF,
		GAMEMSG = 0x2a,
		FILETRANSFER = 0x46,
		VOICECHAT = 0x4A,
		NOTIFY,
		VERIFY,
		P2PFILEXFER,
		PEERTOPEER = 0x4F,	/* Checks if P2P possible */
		WEBCAM,
		AUTHRESP = 0x54,
		LIST,
		AUTH = 0x57,
		ADDBUDDY = 0x83,
		REMBUDDY,
		IGNORECONTACT,	/* > 1, 7, 13 < 1, 66, 13, 0*/
		REJECTCONTACT,
		GROUPRENAME = 0x89, /* > 1, 65(new), 66(0), 67(old) */
		CHATONLINE = 0x96, /* > 109(id), 1, 6(abcde) < 0,1*/
		CHATGOTO,
		CHATJOIN,	/* > 1 104-room 129-1600326591 62-2 */
		CHATLEAVE,
		CHATEXIT = 0x9b,
		CHATLOGOUT = 0xa0,
		CHATPING,
		COMMENT = 0xa8
	}
	
	public enum YahooLoginStatus 
	{
		OK = 0,
		UNAME = 3,
		PASSWD = 13,
		LOCK = 14,
		DUPL = 99,
		SOCK = -1
	}

	public enum YahooStatus 
	{
		AVAILABLE = 0,
		BRB,
		BUSY,
		NOTATHOME,
		NOTATDESK,
		NOTINOFFICE,
		ONPHONE,
		ONVACATION,
		OUTTOLUNCH,
		STEPPEDOUT,
		INVISIBLE = 12,
		CUSTOM = 99,
		IDLE = 999,
		OFFLINE = 0x5a55aa56, /* don't ask */
		NOTIFY = 0x16
	}
	#endregion

	#region [U[Xg֘A
	public class YahooUserGroup
	{
		public string groupname;
		public string[] friendname;
		public YahooUserGroup() { friendname = new string[0]; }
		public void Add(string friend)
		{
			string[] new_friendname = new string[friendname.Length + 1];
			friendname.CopyTo(new_friendname, 0);
			new_friendname[friendname.Length] = friend;
			friendname = new_friendname;
		}
	}

	public class YahooUserList
	{
		public YahooUserGroup[] group;
		public YahooUserList() { group = new YahooUserGroup[0]; }
		public void Add(YahooUserGroup gr)
		{
			YahooUserGroup[] new_group = new YahooUserGroup[group.Length + 1];
			group.CopyTo(new_group, 0);
			new_group[group.Length] = gr;
			group = new_group;
		}
		public void Add(byte[] b)
		{
			// O[v:[U,[U 0x0a
			// O[vƂA[UȂƂɂ͉Ă邩mF
			byte[] tmp = new byte[b.Length];
			int count = 0;
			int state = 0;
			YahooUserGroup g = new YahooUserGroup();
			Encoding enc = Encoding.GetEncoding("shift_jis"); // ȂSJIŜ܂

			for(int i = 0; i < b.Length; i++)
			{
				switch(state)
				{
					case 0:
						if(b[i] == (byte)':')
						{
							g.groupname = enc.GetString(tmp, 0, count);							
							state = 1;
							count = 0;
						}
						else
						{
							tmp[count++] = b[i];
						}
						break;
					case 1:
						if(b[i] == (byte)',')
						{
							g.Add(enc.GetString(tmp, 0, count));
							state = 1;
							count = 0;
						}
						else if(b[i] == 0x0a)
						{
							g.Add(enc.GetString(tmp, 0, count));
							Add(g);	
							g = new YahooUserGroup();
							state = 0;
							count = 0;
						}
						else
						{
							tmp[count++] = b[i];
						}
						break;
				}
			}
			//YahooUtility.PrintUserList(this);
		}
	}
	#endregion

	public interface IYahooMessenger
	{
		void LoginResponse(YahooData yd, int status);
		void GotBuddies(YahooData yd, YahooUserList buddies);
		void GotIgnore(YahooData yd, YahooUserList ignore);
		void StatusChanged(YahooData yd, string name, int status, string msg, bool busy);
		void SystemMessage(YahooData yd, string msg);
		void GotIm(YahooData yd, string from, string msg, int tm, int status, int utf8);
		void TypingNotify(YahooData yd, string from, int status);
		void ErrorNotify(string msg);
		void SendIm(int status, string friend, string msg);
	}

	public class YahooProcess
	{
        private IYahooMessenger owner;
		public string server;
		public int port;
		private TcpClient client;
		private AsyncCallback callback;
		private YahooData person;
		private IAsyncResult asyncresult;
		private byte[] received;
		private byte[] buffer;

		public YahooProcess(IYahooMessenger _owner)
		{
			owner = _owner;
			server = "cs.Yahoo.co.jp";
			port = 5050;
			received = new byte[256];
			buffer = new byte[0];
			client = new TcpClient();
		}
	
		#region ڑ֌W
		private void SocketConnect()
		{
			//System.Diagnostics.Debug.WriteLine("ڑ -> Jn");
			client.Connect(server, port);
			callback = new AsyncCallback(SocketReceive);
			NetworkStream stream = client.GetStream();
			SocketBeginReceive();
		}

		private void SocketClose()
		{
			//System.Diagnostics.Debug.WriteLine("ڑ -> I");
			if(asyncresult == null)
				return;		
			client.Close();
		}

		private void SocketError()
		{
		}

		private void SocketBeginReceive()
		{
			try
			{
				NetworkStream stream = client.GetStream();
				asyncresult = stream.BeginRead(received, 0, received.Length, callback, null);
			} 
			catch
			{
				owner.ErrorNotify("MJn -> s");
			}
		}

		// 񓯊MR[obN֐
		private void SocketReceive(IAsyncResult ar)
		{
			lock(typeof(YahooProcess))
			{
				try
				{
					//System.Diagnostics.Debug.WriteLine("M -> CriticalSection");
			
					NetworkStream stream = client.GetStream();
					int readbyte = stream.EndRead(ar);

					// f[^̍XV
					byte[] newdata = new byte[buffer.Length + readbyte];
					buffer.CopyTo(newdata, 0);
					Buffer.BlockCopy(received, 0, newdata, buffer.Length, readbyte);
					buffer = newdata;
			
					// c̊mF
					if(stream.DataAvailable)
						asyncresult = stream.BeginRead(received, 0, received.Length, callback, ar.AsyncState);
					else
					{
						//System.Diagnostics.Debug.WriteLine("M -> Complete");
						//YahooUtility.DumpByteArray(buffer);

						YahooPacket p;
						int next;
						while(YahooUtility.GetPacket(buffer, out p, out next)) 
						{
							//System.Diagnostics.Debug.WriteLine("M -> pPbg쐬");
							PacketProcess(person, p);

							byte[] new_buffer = new byte[buffer.Length - next];
							Buffer.BlockCopy(buffer, next, new_buffer, 0, buffer.Length - next);
							buffer = new_buffer;
						}
						buffer = new byte[0];
						SocketBeginReceive();
					}
				}
				catch(Exception e)
				{
					owner.ErrorNotify("SocketReceive : " + e.Message);
				}
			}
		}
		#endregion

		#region pPbg쐬
		private YahooPacket CreateLoginPacket(YahooData yd)
		{
            Encoding enc = Encoding.GetEncoding(YahooUtility.MessageEncoding);
			byte[] name = enc.GetBytes(yd.name);
			YahooPacket p = new YahooPacket((short)0x57, 0, 0);
			p.Add(1, name);
			return p;
		}

		private YahooPacket CreateAuthPacket(YahooData yd, string seed)
		{
			Utility.MD5  md5 = new Utility.MD5();
			byte sv = (byte)seed[15];
			sv = (byte)((sv % 8) % 5);
			byte[] b;
			// pX[h̃nbVl擾
			b = Encoding.ASCII.GetBytes(yd.password);
			md5.init();
			md5.update(b, b.Length);
			md5.finalize();
			byte[] pw_hash = md5.digest;
			pw_hash = YahooUtility.to_y64(pw_hash, 16);

			// pX[hmd5_crypt(salt:_2S43d5f)̃nbVl擾
			string crypt = Utility.crypt.md5_crypt(yd.password, "_2S43d5f");
			b = Encoding.ASCII.GetBytes(crypt);
			md5.init();
			md5.update(b, b.Length);
			md5.finalize();
			byte[] crypt_hash = md5.digest;
			crypt_hash = YahooUtility.to_y64(crypt_hash, 16);

			// X|Xl߂
			char checksum;
			byte[] hash_string_p = new byte[49 + yd.name.Length];
			byte[] hash_string_c = new byte[49 + yd.name.Length];
			int pos_p = 0, pos_c = 0;
			switch(sv)
			{
				case 0:
					checksum = seed[seed[7] % 16];
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, (byte)checksum);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, pw_hash);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, yd.name);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, seed);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, (byte)checksum);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, crypt_hash);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, yd.name);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, seed);
					break;
				case 1:
					checksum = seed[seed[9] % 16];
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, (byte)checksum);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, yd.name);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, seed);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, pw_hash);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, (byte)checksum);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, yd.name);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, seed);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, crypt_hash);
					break;
				case 2:
					checksum = seed[seed[15] % 16];
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, (byte)checksum);					
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, seed);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, pw_hash);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, yd.name);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, (byte)checksum);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, seed);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, crypt_hash);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, yd.name);
					break;
				case 3:
					checksum = seed[seed[1] % 16];
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, (byte)checksum);					
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, yd.name);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, pw_hash);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, seed);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, (byte)checksum);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, yd.name);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, crypt_hash);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, seed);
					break;
				case 4:
					checksum = seed[seed[3] % 16];
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, (byte)checksum);					
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, pw_hash);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, seed);
					pos_p = YahooUtility.PutDataToByteArray(hash_string_p, pos_p, yd.name);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, (byte)checksum);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, crypt_hash);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, seed);
					pos_c = YahooUtility.PutDataToByteArray(hash_string_c, pos_c, yd.name);
					break;
			}
			md5.init();
			md5.update(hash_string_p, hash_string_p.Length);
			md5.finalize();
			byte[] result6  = YahooUtility.to_y64(md5.digest, 16);
			md5.init();
			md5.update(hash_string_c, hash_string_c.Length);
			md5.finalize();
			byte[] result96 = YahooUtility.to_y64(md5.digest, 16);
			
			// pPbg쐬
			YahooPacket p = new YahooPacket((short)0x54, 0, 0);
			p.Add(0, Encoding.ASCII.GetBytes(yd.name));
			p.Add(1, Encoding.ASCII.GetBytes(yd.name));
			p.Add(2, new byte[]{ 0x1 });
			p.Add(6, result6);
			p.Add(96, result96);
			return p;
		}

		private YahooPacket CreateStatusChangePacket(YahooData yd, int state, string msg, bool away)
		{
			int current_status;
			short service;

			if(msg != "")
				current_status = (int)YahooStatus.CUSTOM;
			else
				current_status = state;
			if(current_status == (int)YahooStatus.AVAILABLE)
				service = (short)YahooService.ISBACK;
			else
				service = (short)YahooService.ISAWAY;
			YahooPacket p = new YahooPacket(service, current_status, yd.id);
			p.Add(10, Encoding.ASCII.GetBytes(current_status.ToString()));
			if(current_status == (int)YahooStatus.CUSTOM)
			{
                Encoding enc = Encoding.GetEncoding("shift_jis"); // [Jϊ̂SJIS
				p.Add(19, enc.GetBytes(msg));
				p.Add(47, away ? enc.GetBytes("1") : enc.GetBytes("0"));
			}
			return p;
		}

		private YahooPacket CreateIMPacket(YahooData yd, int status, string friend, string msg)
		{
			// status = 0                   : 肪ItCȂ瑗MȂ
			// status = YahooStatus.OFFLINE : FBɑM
			// status = YahooStatus.OFFLINE - 1(`): FBȊOɑM
			YahooPacket p = new YahooPacket((short)YahooService.MESSAGE, status, yd.id);
			p.Add(1, Encoding.ASCII.GetBytes(yd.name));
			p.Add(5, Encoding.ASCII.GetBytes(friend));
            Encoding enc = Encoding.GetEncoding(YahooUtility.MessageEncoding);
			p.Add(14, enc.GetBytes(msg));
			return p;
		}
		#endregion

		#region pPbg
		public void PacketProcess(YahooData yd, YahooPacket packet)
		{
			person = yd;
			switch(packet.service)
			{
				case (short)YahooService.USERSTAT:
				case (short)YahooService.LOGON:
				case (short)YahooService.LOGOFF:
				case (short)YahooService.ISAWAY:
				case (short)YahooService.ISBACK:
				case (short)YahooService.GAMELOGON:
				case (short)YahooService.GAMELOGOFF:
				case (short)YahooService.IDACT:
				case (short)YahooService.IDDEACT:
					ProcessStatus(yd, packet);
					break;
				case (short)YahooService.MESSAGE:
				case (short)YahooService.GAMEMSG:
				case (short)YahooService.SYSMESSAGE:
					ProcessMessage(yd, packet);
					break;
				case (short)YahooService.NEWMAIL:
					ProcessMail(yd, packet);
					break;
				case (short)YahooService.NEWCONTACT:
					ProcessContact(yd, packet);
					break;
				case (short)YahooService.NOTIFY:
					ProcessNotify(yd, packet);
					break;
				case (short)YahooService.VERIFY:
					ProcessVerify(yd, packet);
					break;
				case (short)YahooService.AUTHRESP:
					ProcessAuthResponse(yd, packet);
					break;
				case (short)YahooService.LIST:
					ProcessList(yd, packet);
					break;
				case (short)YahooService.AUTH:
					ProcessAuth(yd, packet);
					break;
				case (short)YahooService.CONFINVITE:
				case (short)YahooService.CONFADDINVITE:
				case (short)YahooService.CONFDECLINE:
				case (short)YahooService.CONFLOGON:
				case (short)YahooService.CONFLOGOFF:
				case (short)YahooService.CONFMSG:
					//yahoo_process_conference(yid, pkt);
					//break;
				case (short)YahooService.CHATONLINE:
				case (short)YahooService.CHATGOTO:
				case (short)YahooService.CHATJOIN:
				case (short)YahooService.CHATLEAVE:
				case (short)YahooService.CHATEXIT:
				case (short)YahooService.CHATLOGOUT:
				case (short)YahooService.CHATPING:
				case (short)YahooService.COMMENT:
					//yahoo_process_chat(yid, pkt);
					//break;
				case (short)YahooService.P2PFILEXFER:
				case (short)YahooService.FILETRANSFER:
					//yahoo_process_filetransfer(yid, pkt);
					//break;
				case (short)YahooService.ADDBUDDY:
					//yahoo_process_buddyadd(yid, pkt);
					//break;
				case (short)YahooService.REMBUDDY:
					//yahoo_process_buddydel(yid, pkt);
					//break;
				case (short)YahooService.IGNORECONTACT:
					//yahoo_process_ignore(yid, pkt);
					//break;
				case (short)YahooService.VOICECHAT:
					//yahoo_process_voicechat(yid, pkt);
					//break;
				case (short)YahooService.WEBCAM:
					//yahoo_process_webcam_key(yid, pkt);
					//break;
				case (short)YahooService.IDLE:
				case (short)YahooService.MAILSTAT:
				case (short)YahooService.CHATINVITE:
				case (short)YahooService.CALENDAR:
				case (short)YahooService.NEWPERSONALMAIL:
				case (short)YahooService.ADDIDENT:
				case (short)YahooService.ADDIGNORE:
				case (short)YahooService.PING:
				case (short)YahooService.GOTGROUPRENAME:
				case (short)YahooService.GROUPRENAME:
				case (short)YahooService.PASSTHROUGH2:
				case (short)YahooService.CHATLOGON:
				case (short)YahooService.CHATLOGOFF:
				case (short)YahooService.CHATMSG:
				case (short)YahooService.REJECTCONTACT:
				case (short)YahooService.PEERTOPEER:
					ProcessUnhandled(packet);
					break;
				default:
					break;
			}
		}

		private void ProcessUnhandled(YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> 舵Ă܂");
            Encoding enc = Encoding.GetEncoding(YahooUtility.MessageEncoding);
			string s;
			for(int i = 0; i < packet.data.Length; i++)
			{
				s = enc.GetString(packet.data[i].data, 0, packet.data[i].data.Length);
				//System.Diagnostics.Debug.WriteLine("0x" + packet.data[i].type.ToString("x3") + " -> " + s);
			}
		}

		private void ProcessVerify(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> Verify");
			try
			{
				SocketConnect();
				YahooPacket p = CreateLoginPacket(yd);
				byte[] data = p.Dump();
				NetworkStream stream = client.GetStream();
				stream.Write(data, 0, data.Length);
			}
			catch
			{
				owner.ErrorNotify("vZX -> Verify -> s");
			}
		}

		private void ProcessAuth(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> F");
			try
			{
				byte[] seed = new byte[0];
				for(int i = 0; i < packet.data.Length; i++)
				{
					if(packet.data[i].type == 94)
						seed = packet.data[i].data;
				}
				YahooPacket p = CreateAuthPacket(yd, Encoding.ASCII.GetString(seed, 0, seed.Length));
				byte[] data = p.Dump();
				NetworkStream stream = client.GetStream();
				stream.Write(data, 0, data.Length);
			}
			catch
			{
				owner.ErrorNotify("vZX -> F -> s");
			}
		}

		private void ProcessAuthResponse(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("F -> ");
			
			int status = 0;
			for(int i = 0; i < packet.data.Length; i++)
			{
				switch(packet.data[i].type)
				{
					case 66:
						status = packet.data[i].data[0];
						break;
				}
			}
			if(packet.status == -1)
			{
				//System.Diagnostics.Debug.WriteLine("F -> s");
				yd.Login = false;
				SocketClose();
				owner.LoginResponse(yd, status);
			}
		}

		private void ProcessList(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> [UXg");
			YahooUserList friend = new YahooUserList();
			YahooUserList ignore = new YahooUserList();

			if(!yd.Login)
			{
				yd.Login = true;
				owner.LoginResponse(yd, (int)YahooLoginStatus.OK);
			}

			for(int i = 0; i < packet.data.Length; i++)
			{
				switch(packet.data[i].type)
				{
					case 87:
						friend.Add(packet.data[i].data);
						break;
					case 88:
						ignore.Add(packet.data[i].data);
						break;
					case 89: // identity???
						break;
					case 59: // cookie???
						break;
				}
			}
			owner.GotBuddies(yd, friend);
			owner.GotIgnore(yd, ignore);
		}
		private void ProcessStatus(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> ԕύX");
			string name = "", msg = "";
			int state = 0, busy = 0, idle = 0, online;
			byte[] b;
            Encoding enc = Encoding.GetEncoding(YahooUtility.MessageEncoding);

			if((packet.service == (short)YahooService.LOGOFF) && (packet.status == -1))
			{
				owner.LoginResponse(yd, (int)YahooLoginStatus.DUPL);
				return ;
			}

			for(int i = 0; i < packet.data.Length; i++)
			{
				b = packet.data[i].data;
				switch(packet.data[i].type)
				{
					case 1:
						if(!yd.Login)
						{
							yd.Login = true;
							owner.LoginResponse(yd, (int)YahooLoginStatus.OK);
						}
						break;
					case 0:
					case 8:
						//System.Diagnostics.Debug.WriteLine(packet.data[i].type.ToString());
						//YahooUtility.DumpByteArray(b);
						break;
					case 7:
						name = Encoding.ASCII.GetString(b, 0, b.Length);
						break;
					case 10:
						state = System.Convert.ToInt32(Encoding.ASCII.GetString(b, 0, b.Length));
						break;
					case 19:
						msg = enc.GetString(b, 0, b.Length);
						break;
					case 47:
						busy = System.Convert.ToInt32(Encoding.ASCII.GetString(b, 0, b.Length));
						break;
					case 137:
						idle = System.Convert.ToInt32(Encoding.ASCII.GetString(b, 0, b.Length));
						break;
					case 13:
						online = System.Convert.ToInt32(Encoding.ASCII.GetString(b, 0, b.Length));
						if((packet.service == (short)YahooService.LOGOFF) || (online == 0))
						{
							owner.StatusChanged(yd, name, (int)YahooStatus.OFFLINE, "", true);
							break;
						}
						if(state == (int)YahooStatus.AVAILABLE)
						{
							owner.StatusChanged(yd, name, state, "", false);
						}
						else if(state == (int)YahooStatus.CUSTOM)
						{
							owner.StatusChanged(yd, name, state, msg, busy != 0);
						}
						else
						{
							owner.StatusChanged(yd, name, state, "", idle != 0);
						}
						break;
					case 11: // ???
					case 17: // ???
						break;
				}
			}
		}

		private void ProcessMessage(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> bZ[W");
			byte[] b;
			string from = "", to = "", msg = "";
			int tm = 0, utf8 = 0, i31 = 0, i32 = 0;
			Encoding enc = Encoding.GetEncoding(YahooUtility.MessageEncoding);
			for(int i = 0; i < packet.data.Length; i++)
			{
				b = packet.data[i].data;
				switch(packet.data[i].type)
				{
					case 1:
					case 4:
						from = enc.GetString(b, 0, b.Length);
						break;
					case 5:
						to = enc.GetString(b, 0, b.Length);
						break;
					case 15:
						tm = System.Convert.ToInt32(enc.GetString(b, 0, b.Length));
						break;
					case 97:
						utf8 = System.Convert.ToInt32(enc.GetString(b, 0, b.Length));
						break;
					case 14:
					case 16:
						msg = enc.GetString(b, 0, b.Length);
						break;
					case 31: // dv
						// ɏ܂
						i31 = System.Convert.ToInt32(enc.GetString(b, 0, b.Length));
						break;
					case 32:
						i32 = System.Convert.ToInt32(enc.GetString(b, 0, b.Length));
						break;
				}
			}
			if (packet.service == (short)YahooService.SYSMESSAGE) 
			{
				owner.SystemMessage(yd, msg);
			} 
			else if (packet.status <= 2 || packet.status == 5) 
			{
				owner.GotIm(yd, from, YahooUtility.GetMessageString(msg), tm, packet.status, utf8);
			} 
			else if (packet.status == -1) 
			{
				owner.ErrorNotify("ProcessMsg : " +msg);
			}
		}

		private void ProcessNotify(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> ʒm");
			byte[] b;
			string from = "", msg = "", ind = "";
			int stat = 0;
            Encoding enc = Encoding.GetEncoding(YahooUtility.MessageEncoding);

			for(int i = 0; i < packet.data.Length; i++)
			{
				b = packet.data[i].data;
				switch(packet.data[i].type)
				{
					case 4:
						from = enc.GetString(b, 0, b.Length);
						break;
					case 49:
						msg = enc.GetString(b, 0, b.Length);
						break;
					case 13:
						stat = System.Convert.ToInt32(enc.GetString(b, 0, b.Length));
						break;
					case 14:
						ind = enc.GetString(b, 0, b.Length);
						break;
				}
			}

			if(msg == "")
				return;

			if(msg == "TYPING")
			{
				owner.TypingNotify(yd, from, stat);
			}
			else if(msg == "GAME")
			{
				// 
			}
			else if(msg == "WEBCAMINVITE")
			{
				// 
			}
			else
			{
				// [H
			}
		}

		private void ProcessMail(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> [");
			// 
		}

		private void ProcessContact(YahooData yd, YahooPacket packet)
		{
			//System.Diagnostics.Debug.WriteLine("vZX -> R^Ng");
			// 
		}
		#endregion

		#region [U
		public void ChangeStatus(int state, string msg, bool away)
		{
			//System.Diagnostics.Debug.WriteLine("ChangeStatus");
			if(person == null)
				return;
			try
			{
				YahooPacket p = this.CreateStatusChangePacket(person, state, msg, away);
				NetworkStream stream = client.GetStream();
				byte[] data = p.Dump();
				stream.Write(data, 0, data.Length);
				owner.StatusChanged(person, person.name, state, msg, away);
			}
			catch
			{
				owner.ErrorNotify("ԕύX -> s");
			}
		}

		public void SendIM(int status, string friend, string msg)
		{
			//System.Diagnostics.Debug.WriteLine("SendIM");
			if(person == null)
				return;
			try
			{
				YahooPacket p = CreateIMPacket(person, status, friend, msg);
				NetworkStream stream = client.GetStream();
				byte[] data = p.Dump();
				stream.Write(data, 0, data.Length);
			}
			catch
			{
				owner.ErrorNotify("bZ[WM -> s");
			}
		}

		public void Disconnect()
		{
			SocketClose();
		}
		#endregion
	}

	#region YahooProtocolp[eBeB֐Q
	public class YahooUtility
	{
        public const string MessageEncoding = "utf-8";

		#region bZ[Ŵ݂o
		public static string GetMessageString(string msg)
		{
			bool isLoop = true;
			int begin, end;
			while(isLoop)
			{
				isLoop = false;
				// <font xxx >̓JbgB
				begin = msg.IndexOf("<font ");
				if(begin != -1)
				{
					end = msg.IndexOf(">", begin);
					if(end != -1)
					{
						msg = msg.Remove(begin, end - begin + 1);
						isLoop = true;
					}
				}
				// AΐAAFύXJbg
				// ́[ xxxx m@(0x1b 0x5b xxxx 0x6d)
				// FύX̍ہAxxxx͂₽璷Ƃ̂Œ
				begin = msg.IndexOf("\x001b\x005b");
				if(begin != -1)
				{
					end = msg.IndexOf("m", begin);
					if(end != -1)
					{
						msg = msg.Remove(begin, end - begin + 1);
						isLoop = true;
					}
				}
			}
			return msg;
		}
		#endregion

		#region ԕ𓾂֐
		public static string GetStatusString(int status, string msg)
		{
			switch(status)
			{
				case (int)YahooStatus.AVAILABLE:
					return "( gp )";
				case (int)YahooStatus.BUSY:
					return "( ܑZ )";
				case (int)YahooStatus.INVISIBLE:
					return "( ȂtĂĂ )";
				case (int)YahooStatus.IDLE:
					return "( ACh )";
				case (int)YahooStatus.NOTATDESK:
					return "( Ȃ͂Ă܂ )";
				case (int)YahooStatus.NOTATHOME:
					return "( Ƃɂ܂ )";
				case (int)YahooStatus.NOTINOFFICE:
					return "( ItBXɂ܂ )";
				case (int)YahooStatus.OFFLINE:
					return " [ offline ]";
				case (int)YahooStatus.ONPHONE:
					return "( db )";
				case (int)YahooStatus.ONVACATION:
					return "( xɒ )";
				case (int)YahooStatus.OUTTOLUNCH:
					return "( H )";
				case (int)YahooStatus.STEPPEDOUT:
					return "( A )";
				case (int)YahooStatus.CUSTOM:
					return "(" + msg + ")";
				case (int)YahooStatus.BRB:
					return "( ߂܂ )";
				case (int)YahooStatus.NOTIFY: // ????
					return "( Notify )";
				default:
					return "( " + status.ToString() + " )";
			}
		}
		#endregion

		#region o͊֐
		// oCgo͂֐
		public static void DumpByteArray(byte[] b)
		{
			string s;
			int l = 0;

			System.Diagnostics.Debug.WriteLine("");
			System.Diagnostics.Debug.Write("       ");
			for(int i = 0; i < 16; i++)
				System.Diagnostics.Debug.Write("+" + i.ToString("x") + " ");
			System.Diagnostics.Debug.WriteLine("");
			
			int cnt = ((b.Length - 1) / 16 + 1) * 16;
			for(int i = 0; i < cnt; i++)
			{
				if(i % 16 == 0)
					System.Diagnostics.Debug.Write("0x" + i.ToString("x4") + " ");
				
				if(i < b.Length)
				{
					System.Diagnostics.Debug.Write(b[i].ToString("x2") + " ");
					l++;
				}
				else
					System.Diagnostics.Debug.Write("   ");

				if(i % 16 == 15)
				{
					s = Encoding.ASCII.GetString(b, (i / 16) * 16, l);
					System.Diagnostics.Debug.Write(s);
					System.Diagnostics.Debug.WriteLine(" ");
					l = 0;
				}
			}
		}

		public static void PrintUserList(YahooUserList ul)
		{
			System.Diagnostics.Debug.WriteLine("UserList");
			for(int i = 0; i < ul.group.Length; i++)
			{
				System.Diagnostics.Debug.WriteLine(ul.group[i].groupname);
				for(int j = 0; j < ul.group[i].friendname.Length; j++)
				{
					System.Diagnostics.Debug.WriteLine("   " + ul.group[i].friendname[j]);
				}
			}
		}
		#endregion

		#region oCg񂩂YahooPacket𓾂֐
		public static bool GetPacket(byte[] b, out YahooPacket p, out int next)
		{
			p = new YahooPacket();
			next = -1;
			// wb_
			// signaturemF
			if(b.Length < 4)
				return false;
			if(!((b[0] == (byte)'Y') && (b[1] == (byte)'M') && (b[2] == (byte)'S') && (b[3] == (byte)'G')))
				return false;

			int length = (short)((b[8] << 8) + b[9]);
			p.service = (short)((b[10] << 8) + b[11]);
			p.status = (int)((b[12] << 24) + (b[13] << 16) + (b[14] << 8) + b[15]);
			p.id = (int)((b[16] << 24) + (b[17] << 16) + (b[18] << 8) + b[19]);
			next = length + 20;

			// f[^
			YahooParameter param = new YahooParameter();
			byte[] tmp = new byte[length];
			int ps = 0;    // type or data?
			int count = 0;
			int state = 0;
			for(int i = 20; i < length + 20; i++)
			{
				switch(state)
				{
					case 0:
						tmp[count++] = b[i];
						if(b[i] == 0xc0)
							state = 1;
						break;
					case 1:
						tmp[count++] = b[i];
						if(b[i] == 0x80)
						{
							
							if(ps == 0)
							{
								param.type = 0;
								int shift = 1;
								for(int j = count - 3; j >= 0; j--)
								{
									param.type += ((tmp[j] - 0x30) * (shift));
									shift *= 10;
								}
								ps = 1;
							}
							else
							{
								param.data = new byte[count - 2];
								Buffer.BlockCopy(tmp, 0, param.data, 0, count - 2);
								p.Add(param);
								param = new YahooParameter();
								ps = 0;
							}
							count = 0;
						}
						state = 0;
						break;
				}
			}
			return true;
		}
		#endregion

		#region YahoobZWpBase64ǂϊ
		public static string base64string = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" + 
			"abcdefghijklmnopqrstuvwxyz" +
			"0123456789._";		
		/* raw bytes in quasi-big-endian order to base 64 string */
		public static byte[] to_y64(byte[] input, int inlen)
		{
			byte[] result = new byte[((inlen + 2) / 3) * 4];
			byte[] base64digits = Encoding.ASCII.GetBytes(base64string);
			byte[] bar = new byte[1] {(byte)'-'};
			int inOffset = 0;	
			int outOffset = 0;
			for (; inlen >= 3; inlen -= 3)
			{ 
				Buffer.BlockCopy(base64digits,   input[inOffset    ] >> 2, result, outOffset++, 1);
				Buffer.BlockCopy(base64digits, ((input[inOffset    ] << 4) & 0x30) | (input[inOffset + 1] >> 4), result, outOffset++, 1);
				Buffer.BlockCopy(base64digits, ((input[inOffset + 1] << 2) & 0x3c) | (input[inOffset + 2] >> 6), result, outOffset++, 1);
				Buffer.BlockCopy(base64digits,   input[inOffset + 2] & 0x3f, result, outOffset++, 1);
				inOffset += 3;
			}
			
			if (inlen > 0)
			{
				byte fragment;

				Buffer.BlockCopy(base64digits,   input[inOffset] >> 2, result, outOffset++, 1);
				fragment = (byte)((input[inOffset] << 4) & 0x30);
				if (inlen > 1)
					fragment |= (byte)(input[inOffset + 1] >> 4);
				Buffer.BlockCopy(base64digits,   fragment, result, outOffset++, 1);
				if(inlen < 2)
					Buffer.BlockCopy(bar, 0, result, outOffset++, 1);
				else
					Buffer.BlockCopy(base64digits, (input[inOffset + 1] << 2) & 0x3c, result, outOffset++, 1);
				Buffer.BlockCopy(bar, 0, result, outOffset++, 1);
			}
			return result;
		}
		#endregion

		#region oCg𐔒lɕϊ
		// WBitComverterƒȂoCgnƂÔ
		public static int ByteArrayToInt(byte[] b, int index) { return ByteArrayToInt(b, index, false); }
		public static int ByteArrayToInt(byte[] b, int index, bool bigendian)
		{
			int result = 0;
			int len = b.Length - index;
			if(len <= 0)
				return 0;
			if(len > 4)
				len = 4;
			for(int i = 0; i < len; i++)
			{
				if(bigendian)
					result += (int)(b[index + i] << (8 * (len - i - 1)));
				else
					result += (int)(b[index + i] << (8 * i));
			}
			return result;
		}
		#endregion
		
		#region oCgz֖ߍފ֐Q
		public static int PutDataToByteArray(byte[] b, int offset, int data)
		{
			b[offset + 3] = (byte)((data >> 24) & 0xFF);
			b[offset + 2] = (byte)((data >> 16) & 0xFF);
			b[offset + 1] = (byte)((data >>  8) & 0xFF);
			b[offset    ] = (byte)((data      ) & 0xFF);
			return offset + 4;
		}
		
		public static  int PutDataToByteArray(byte[] b, int offset, short data)
		{
			b[offset + 1] = (byte)((data >>  8) & 0xFF);
			b[offset    ] = (byte)((data      ) & 0xFF);
			return offset + 2;
		}
		
		public static  int PutDataToByteArray(byte[] b, int offset, byte data)
		{
			b[offset] = data;
			return offset + 1;
		}

		// ASCIIR[hŕϊ
		public static int PutDataToByteArray(byte[] b, int offset, string s)
		{
			byte[] bs = Encoding.ASCII.GetBytes(s);
			for(int i = 0; i < bs.Length; i++)
				b[offset + i] = bs[i];
			return offset + bs.Length;
		}

		public static int PutDataToByteArray(byte[] b, int offset, byte[] data)
		{
			for(int i = 0; i < data.Length; i++)
				b[offset + i] = data[i];
			return offset + data.Length;
		}

		#endregion

	}
	#endregion
}
